/* -*- C++ -*- 
 * mu_filterqueue.h
 * @(#) header for smart disk-filtered queue classes
 *
 * Copyright (C) 2001 - 2003 by Giuseppe Della Penna              
 * Universita' degli Studi di L'Aquila
 * 
 * Code fragments in this file are from the original Murphi Relese 3.1 
 * Copyright (C) 1992 - 1999 by the Board of Trustees of              
 * Leland Stanford Junior University.
 *
 * Read the file "license" distributed with these sources, or call
 * Murphi with the -l switch for additional information.
 * 
 */

/* 
 * Original Author: G. Della Penna
 * Created: 01 Oct 2001
 * Released: 30 Jan 2003
 *       
 */

#ifndef _FILTERQUEUE_
#define _FILTERQUEUE_

#define NEUTRAL_ZONE_HI_BOUND 0.5
#define NEUTRAL_ZONE_LO_BOUND 0.3

#define MIN_FILTER_ALPHA 0.1
#define INITIAL_ALPHA 0.5
#define CRITICAL_ALPHA 0.95
#define ALPHA_DELAY_FACTOR 1

#define BLOCK_MIN_SIZE 10000
#define BLOCK_MAX_NUMBER 100
#define BLOCK_SIZE_STEP 50

#define CALIBRATION_FREQ 10

#define AREA0_PERCENT_SIZE 0.40
#define AREA1_PERCENT_SIZE 0.30
#define AREA0_MAX_PROB 0.4
#define AREA1_MAX_PROB 0.6

#define AGING 2.0

/***********************************************
  Pair containing a state and its hash signature.
  Used by the unchecked queue.
 ***********************************************/
struct state_and_signature_pair {
    state _state;
    unsigned long _signature[2];

        state_and_signature_pair(state* s, unsigned long c1, unsigned long c2) 
    {_state = *s; _signature[0]=c1; _signature[1]=c2;}

        state_and_signature_pair() { };
};

/**************************************************
  Queue containing states and their hash signature.
  (The "unchecked" queue)
 **************************************************/
class state_and_signature_queue
{
protected:
  state_and_signature_pair* stateArray;   /* The actual array. */
  const unsigned long max_active_states;  /* max size of queue */
  unsigned long front;                    /* index of first active state. */
  unsigned long rear;                     /* index of next free slot. */

  unsigned long global_rear,global_front;
  unsigned long num_elts_head,num_elts_tail;
  unsigned long head_begin,tail_begin;
  unsigned long head_size,tail_size;

  /* Disk paging files */
  splitFile *paging_file_top, *paging_file_bottom;
  
public:
  // initializers
  state_and_signature_queue( unsigned long mas );

  // destructor
  virtual ~state_and_signature_queue();
  
  // information interface
  inline unsigned long MaxElts( void ) { return max_active_states; }
  unsigned long NumElts( void ) { return num_elts_head + num_elts_tail + global_rear + global_front; }
  inline bool isempty( void )  { return num_elts_head + num_elts_tail + global_rear + global_front == 0; }

  inline static int BytesForOneState( void );     

  // storing and removing elements
  virtual void enqueue( state_and_signature_pair* &e );
  virtual state_and_signature_pair* dequeue( void );
  virtual state_and_signature_pair* top( void );
  
  virtual void ReclaimFreeSpace();
  virtual void QueueEmptyFault();

  virtual unsigned NextRuleToTry()   
  {
    Error.Notrace("Internal: Getting next rule to try from a state queue instead of a state stack.");
    return 0;
  }
  virtual void NextRuleToTry(unsigned r)
  {
    Error.Notrace("Internal: Setting next rule to try from a state queue instead of a state stack.");
  }

    // printing routine
  void Print( void );
  virtual void print_capacity( void )
  {
    cout << "\t* Capacity in queue for breadth-first search: "
     << max_active_states << " states.\n"
     << "\t   * Change the constant gPercentActiveStates in mu_prolog.inc\n"
         << "\t     to increase this, if necessary.\n"; 
  }

};



/****************************************
  The disk-filtered state queue.
  (The "checked" queue)
 ****************************************/
class state_filtering_queue
{
#ifdef HASHC
  typedef unsigned long Unsigned32; 
#endif


protected:
  state* stateArray;                      /* The actual array. */
  const unsigned long max_active_states;  /* max size of queue */
  unsigned long front;                    /* index of first active state. */
  unsigned long rear;                     /* index of next free slot. */

  unsigned long global_rear,global_front;
  unsigned long num_elts_head,num_elts_tail;
  unsigned long head_begin,tail_begin;
  unsigned long head_size,tail_size;

  unsigned long history_len;
  double filter_alpha;

  /* Disk paging files */
  splitFile *paging_file_top, *paging_file_bottom;

  state_and_signature_queue *uncheckedQueue;
  splitFile *history_file;

  Unsigned32 *filterTable;
  bool hashtable_lookup(unsigned long, unsigned long, bool, bool);
  unsigned long signature_hash(unsigned long t1, unsigned long t2);
  dynBitVec *Full,*Removed;

  unsigned long hashtable_size;
  unsigned long hashtable_num_elts;

  bool enqueue_checked( state* e );
  state* dequeue_checked( void );
  state* top_checked( void );

  bool enqueue_unchecked( state* &e );
  state_and_signature_pair* dequeue_unchecked( void );
  
  char *QueryReadFront();
  bool QueryCalibration();
  void Calibrate(unsigned long deleted_a, unsigned long deleted_b);
  unsigned long FilterTable(unsigned long states, unsigned long *_totread = NULL);

  double SelectionP(unsigned long block);

  void DefineBlocks();
  void CheckTable_full();
  void CheckTable_simple();

  randomGen random;
  unsigned long history_blocks;
  unsigned long history_block_len;
  unsigned long last_deleted_states;

  bool hashtable_first_full;

public:
  // initializers
  state_filtering_queue( unsigned long mas );

  // destructor
  virtual ~state_filtering_queue();
  
  // information interface
  inline unsigned long MaxElts( void ) { return max_active_states; } //sarebbe meglio dire TANTI...
  unsigned long NumElts( void ) { return num_elts_head + num_elts_tail + global_rear + global_front; }
  unsigned long HistoryLength( void ) { return history_len; }
  double DiskAlpha( void ) { return filter_alpha; }
  inline bool isempty( void );

  //to avoid problems with table lookup, declare it full a bit before it is really full
  //inline bool tableisfull( void ) {return (hashtable_num_elts >= hashtable_size);}
  inline bool tableisfull( void ) {return (hashtable_num_elts >= hashtable_size - 3);}
  
  inline unsigned long deletedstates() {
      unsigned long temp = last_deleted_states;
      last_deleted_states=0;
      return temp;
  }
  
  inline static int BytesForOneState_checked( void ); 
  inline static int BytesForOneState_unchecked( void ); 
  inline static int BytesForOneState_hashtable( void ); 
  
  // storing and removing elements
  virtual bool enqueue( state* &e ) {return enqueue_unchecked(e);}
  virtual state* dequeue( void ) {return dequeue_checked();}
  virtual state* top( void ) {return top_checked();}
  
  virtual void ReclaimFreeSpace();
  virtual void QueueEmptyFault();

  inline void CheckTable();

  virtual unsigned NextRuleToTry()  
  {
    Error.Notrace("Internal: Getting next rule to try from a state queue instead of a state stack.");
    return 0;
  }
  virtual void NextRuleToTry(unsigned r)
  {
    Error.Notrace("Internal: Setting next rule to try from a state queue instead of a state stack.");
  }

  // printing routine
  void Print( void );
  virtual void print_capacity( void )
  {
    cout << "\t* Capacity in queue for breadth-first search:\n"
     << "\t  -" << max_active_states << " checked (old) states.\n"
     << "\t  -" << (int)max_active_states << " unchecked (new) states.\n"
     << "\t  -" << (int)(max_active_states / gPercentActiveStates) << " states in the new state buffer.\n"
     << "\t   * Change the constant gPercentActiveStates in mu_prolog.inc\n"
         << "\t     to increase this, if necessary.\n"; 
  }
};

#endif


/********************
  $Log: mu_filterqueue.h,v $

  Revision 2.0  2003/01/10 17:06:12  giuseppe 
  disk extension - final release

  Revision 1.2  2001/10/28 21:50:00  giuseppe 
  disk algorithm complete - start testing

  Revision 1.1  2001/10/01 18:41:14  giuseppe 
  Initial revision

********************/
