/* -*- C++ -*- 
 * splitFile.C
 * @(#) class for handling VERY large files
 *
 * Copyright (C) 2001 - 2003 by Giuseppe Della Penna              
 * Universita' degli Studi di L'Aquila
 * 
 * Read the file "license" distributed with these sources, or call
 * Murphi with the -l switch for additional information.
 * 
 */

/* 
 * Original Author: G. Della Penna
 * Created: 01 Oct 2001
 * Released: 30 Jan 2003
 *       
 */

/*
 * This file handling class was designed to fix a problem with
 * Linux EXT2 filesystem, that does not handle by default files
 * larger that 1Gb. This problem could also appear on other
 * filesystems, so I decided to write a file class that handles
 * very large "virtual" files storing them on many smaller files.
 * The splitFile class has methods to handle all the normal file
 * operations. Moreover, all the disk-access routines specify
 * lengths (i.e. number of bytes to read, position to seek...)
 * using two unsigned long integers: size and count. The number
 * of bytes requested, e.g. the position to seek to in the file,
 * is size*count.
 *
 * WARNING: the class has been tested in disk murphi only.
 * Some minor functions can still contain bugs.
 * 
 * WARNING: reading a file with a size paramemter different
 * from that used to write may lead to strange errors...
 */
 
#define min(a,b) (((a)<(b))?(a):(b))
#define max(a,b) (((a)>(b))?(a):(b))

//Max # of parts handled by the splitFile class
//Adjust to match your needs
#define MAX_PARTS 100

#include "splitFile.h"

/*
 * Creates (but not opens) a spliFile.
 * The file will be subdivided in several files,
 * each with a length <= slitsize.
 * If permanent is false, the split file will be
 * deleted when closed.
 *
 */
splitFile::splitFile(unsigned long splitsize, bool permanent) :
    _splitsize(splitsize),
        maxpart(-1),
        curpart(-1),
        _eof(true),
        _permanent(permanent)

{
    if (sizeof(char) != 1) {
        fprintf(stderr,"sizeof(char) != 8");
        exit(10);
    }

    memset(parts,0,sizeof(FILE*)*MAX_PARTS);
}

splitFile::~splitFile()
{
    close();
}

void splitFile::CreateNextPart()
{
    //warning: no check for too many parts created (overflow in arrays parts and partsize) 
    char filename[1024];
    sprintf(filename,"%s.%02d",_filename,maxpart+1);

    if ((parts[++maxpart] = fopen(filename,_access))==NULL) {
        fprintf(stderr,"impossibile creare la nuova sezione del file: %s",filename);
        exit(10);
    }
    curpart=maxpart;
    partsize[curpart] = 0;
}

bool splitFile::OpenNextPart()
{
    //warning: no check for too many parts created (overflow in arrays parts and partsize) 
    char filename[1024];
    sprintf(filename,"%s.%02d",_filename,maxpart+1);

    FILE *dummy;

    if ((dummy=fopen(filename,"r"))==NULL) return false;
    fseek(dummy,0,SEEK_END);
    partsize[curpart]=ftell(dummy);
    fclose(dummy);

    if ((parts[++maxpart] = fopen(filename,_access))==NULL) {
        fprintf(stderr,"impossibile aprire la nuova sezione del file: %s",filename);
        exit(10);
    }
    curpart=maxpart;
    
    return true;
}


bool splitFile::GotoNextPart(bool write_mode)
{
    if (parts[curpart+1] == NULL) {
        if (write_mode) CreateNextPart();
        else return false;
    } else {
        curpart+=1;
        fseek(parts[curpart],0,SEEK_SET);
    }
    return true;
}


unsigned long splitFile::curfilepos()
{
    return ftell(parts[curpart]);
}


unsigned long splitFile::getpartsize(int npart)
{
    /*
    if (npart == curpart)
        return max(ftell(parts[curpart]),partsize[curpart]));
    else 
    */
    return partsize[npart];
}


bool splitFile::open(char *filename, char* access)
{
    strcpy(_access,access);
    strcpy(_filename,filename);
    if (strchr(access,'r')) while (OpenNextPart());
    if (maxpart<0) CreateNextPart();
    curpart=0;

    return true;
}

void splitFile::close()
{
    //all the file parts are opened and registered in the parts array
    char filename[1024];

    for(int i=0; i<=maxpart; i++) {
        if(parts[i]!=NULL) {
            fclose(parts[i]);
            if (!_permanent) {
                sprintf(filename,"%s.%02d",_filename,i);
                remove(filename);
            }
        }
    }
    
    memset(parts,0,sizeof(FILE*)*MAX_PARTS);
    maxpart = -1;
    curpart = -1;
    _eof = true;
}

/*
 * splitFile read and write routines are identical to the
 * corresponding FILE* versions.
 * 
 */
unsigned long splitFile::write(void * buffer, unsigned long size, unsigned long count)
{
    unsigned long inthispart, towrite,totalwritten=0;
    unsigned long pp=0;
    
    _eof=false;

    while(count>0) {
        inthispart= (_splitsize-curfilepos())/size;
        towrite = min(count,inthispart);
        fwrite(((char*)buffer)+pp,size,towrite,parts[curpart]);
        count-=towrite;
        totalwritten+=towrite;
        pp+=towrite*size;
        if (count>0) {
            partsize[curpart] = curfilepos(); 
            GotoNextPart(true);
        }
        if (ftell(parts[curpart])>partsize[curpart])
            partsize[curpart] = ftell(parts[curpart]);
    }

    return totalwritten;
}

unsigned long splitFile::read(void * buffer, unsigned long size, unsigned long count)
{
    unsigned long inthispart, toread, read, totalread=0;
    unsigned long pp=0;

    while(count>0) {
        inthispart= (_splitsize-curfilepos())/size;
        toread = min(count,inthispart);
        read = fread(((char*)buffer)+pp,size,toread,parts[curpart]);
        count-=read;
        totalread+=read;
        pp+=read*size;
        if (count>0) {
            partsize[curpart] = curfilepos(); 
            if (read==toread) {
                if (!GotoNextPart(false)) {
                    _eof=true;
                    break;  
                }
            } else {
                _eof=true;
                break;  
            }
            
        }
        if (ftell(parts[curpart])>partsize[curpart])
            partsize[curpart] = ftell(parts[curpart]);
    }

    return totalread;
}

/*
 * splitFile seek takes two parameters to specify
 * the position - like read and write.
 * The two-argument version assumes that the size is 1 byte
 * (so actually it works like the FILE* seek function)
 * WARNING: SEEK_CUR currently not supported.
 *
 */
void splitFile::seek(unsigned long count, int from)
{
    seek(count,1,from);
}

void splitFile::seek(unsigned long count, unsigned long size, int from)
{
    int i;
    
    if (from == SEEK_SET) {
        for (i = 0; i<=maxpart; i++) {
//            fprintf(stderr, "\nseek(): looking in part %d, %x (max=%d)",i,parts[i],maxpart);
            if (getpartsize(i)/size > count) {
                break;
            } else {
                count -= (getpartsize(i)/size);
            }
        }
        curpart=min(i,maxpart);
//        fprintf(stderr, "\nseek(): selected part %d, %x: seeking...",curpart,parts[curpart]);
        if (curpart>=0) fseek(parts[curpart],count*size,SEEK_SET);
//        fprintf(stderr, "\nseek(): seeking to part %d, pos %ul",curpart,count*size);
    } else if (from ==  SEEK_END) {
        for (i = maxpart; i>=0; i--) {
            if (getpartsize(i)/size > count) {
                break;
            } else {
                count -= (getpartsize(i)/size);
            }
        }
        curpart=max(i,min(0,maxpart));
        if (curpart>=0) fseek(parts[curpart],count*size,SEEK_END);
    } else {
        fprintf(stderr,"SEEK_CUR not valid in splitFile");
        exit(10);
    }

    _eof=false;
}

unsigned long splitFile::tell()
{
    unsigned long total=0;
    for (int i = 0; i<curpart; i++) 
         total += getpartsize(i);
    
    return total+curfilepos();
}


bool splitFile::eof()
{
    return _eof;    
}

/********************
  $Log: mu_splitFile.C,v $

  Revision 2.0  2003/01/10 17:06:12  giuseppe 
  disk extension - final release

  Revision 1.2  2001/10/28 21:50:00  giuseppe 
  disk algorithm complete - start testing

  Revision 1.1  2001/10/01 18:41:14  giuseppe 
  Initial revision

********************/
